#include "transfers.h"
#include "command_buffers.h"

#include "runtime/config.h"
#include "runtime/mem_arena.h"
#include "runtime/threading.h"

#include <stdbool.h>

RT_CVAR_I(rt_VkTransferSlotCount,
          "Number of available transfer slots per frame. Default: 512",
          512);

/* This is a temporary solution. We probably should keep a pool of buffers
 * to avoid re-creating the buffers all the time. */

typedef struct {
    VkBuffer buffer;
    VmaAllocation allocation;
    bool requires_flush;
} rt_transfer_buffer;

typedef struct {
    rt_transfer_buffer tbuf;
    VkFence fence;
    VkSemaphore ownership_transfer;
} rt_transfer;

static rt_transfer *_transfers;
static uint32_t _transfer_count;
static rt_mutex *_transfer_lock;

static rt_transfer_buffer AcquireTransferBuffer(size_t size) {
    rt_transfer_buffer tbuf = {VK_NULL_HANDLE};

    VkBufferCreateInfo buffer_info = {
        .sType       = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
        .size        = size,
        .usage       = VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
        .sharingMode = VK_SHARING_MODE_EXCLUSIVE,
    };
    VmaAllocationCreateInfo alloc_info = {
        .usage = VMA_MEMORY_USAGE_AUTO,
        .flags = VMA_ALLOCATION_CREATE_HOST_ACCESS_SEQUENTIAL_WRITE_BIT,
    };

    if (vmaCreateBuffer(g_gpu.allocator,
                        &buffer_info,
                        &alloc_info,
                        &tbuf.buffer,
                        &tbuf.allocation,
                        NULL) == VK_SUCCESS) {
        VkMemoryPropertyFlags props;
        vmaGetAllocationMemoryProperties(g_gpu.allocator, tbuf.allocation, &props);
        tbuf.requires_flush = (props & VK_MEMORY_PROPERTY_HOST_COHERENT_BIT) == 0;
    }
    return tbuf;
}

static void ReturnTransferBuffer(rt_transfer_buffer buffer) {
    vmaDestroyBuffer(g_gpu.allocator, buffer.buffer, buffer.allocation);
}

static void CopyToTransferBuffer(rt_transfer_buffer buffer, const void *data, size_t n) {
    void *tmem = NULL;
    vmaMapMemory(g_gpu.allocator, buffer.allocation, &tmem);
    RT_ASSERT(tmem, "Transfer Buffer memory must be mappable.");
    memcpy(tmem, data, n);
    vmaUnmapMemory(g_gpu.allocator, buffer.allocation);
    if (buffer.requires_flush)
        vmaFlushAllocation(g_gpu.allocator, buffer.allocation, 0, n);
}

rt_result InitializeTransfers(void) {
    _transfer_lock = rtCreateMutex();
    if (!_transfer_lock)
        return RT_UNKNOWN_ERROR;
    _transfers = calloc((size_t)rt_VkTransferSlotCount.i, sizeof(rt_transfer));
    if (!_transfers) {
        rtDestroyMutex(_transfer_lock);
        return RT_OUT_OF_MEMORY;
    }
    _transfer_count = 0;
    return RT_SUCCESS;
}

void ShutdownTransfers(void) {
    rtDestroyMutex(_transfer_lock);
    for (int i = 0; i < rt_VkTransferSlotCount.i; ++i) {
        if (_transfers[i].fence)
            vkDestroyFence(g_gpu.device, _transfers[i].fence, g_gpu.alloc_cb);
    }
    free(_transfers);
}

#define TRANSFER_FAILED     -1
#define TRANSFER_NOT_NEEDED 0
#define TRANSFER_STARTED    1

static int AcquireBufferOwnership(rt_transfer *transfer,
                                  VkBuffer buffer,
                                  rt_gpu_queue current_owner,
                                  VkCommandBuffer transfer_cmd) {
    if (!transfer->ownership_transfer) {
        VkSemaphoreCreateInfo sem_info = {
            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
        };
        if (vkCreateSemaphore(g_gpu.device,
                              &sem_info,
                              g_gpu.alloc_cb,
                              &transfer->ownership_transfer) != VK_SUCCESS) {
            rtReportError("vk", "Failed to create an ownership transfer semaphore.");
            return TRANSFER_FAILED;
        }
    }

    uint32_t src_family = rtGetQueueFamily(current_owner);
    uint32_t dst_family = rtGetQueueFamily(RT_TRANSFER_QUEUE);
    if (src_family == dst_family)
        return TRANSFER_NOT_NEEDED;

    VkCommandBuffer cmd                 = rtAllocSingleCommandBuffer(current_owner);
    VkCommandBufferBeginInfo begin_info = {.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
                                           .flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT};
    vkBeginCommandBuffer(cmd, &begin_info);
    VkBufferMemoryBarrier2 release_barrier = {
        .sType               = VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER_2,
        .buffer              = buffer,
        .offset              = 0,
        .size                = VK_WHOLE_SIZE,
        .srcStageMask        = VK_PIPELINE_STAGE_2_BOTTOM_OF_PIPE_BIT,
        .srcAccessMask       = 0,
        .srcQueueFamilyIndex = src_family,
        .dstQueueFamilyIndex = dst_family,
    };
    VkDependencyInfo dep = {.sType                    = VK_STRUCTURE_TYPE_DEPENDENCY_INFO,
                            .pBufferMemoryBarriers    = &release_barrier,
                            .bufferMemoryBarrierCount = 1};
    vkCmdPipelineBarrier2(cmd, &dep);
    vkEndCommandBuffer(cmd);

    VkBufferMemoryBarrier2 acquire_barrier = {
        .sType               = VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER_2,
        .buffer              = buffer,
        .offset              = 0,
        .size                = VK_WHOLE_SIZE,
        .dstStageMask        = VK_PIPELINE_STAGE_2_TRANSFER_BIT,
        .dstAccessMask       = VK_ACCESS_MEMORY_WRITE_BIT,
        .srcQueueFamilyIndex = src_family,
        .dstQueueFamilyIndex = dst_family,
    };
    VkDependencyInfo dep2 = {.sType                    = VK_STRUCTURE_TYPE_DEPENDENCY_INFO,
                             .pBufferMemoryBarriers    = &acquire_barrier,
                             .bufferMemoryBarrierCount = 1};
    vkCmdPipelineBarrier2(transfer_cmd, &dep2);
    
    /* Only transfer the ownership when the frame is finished */
    VkSemaphore wait_semaphore = VK_NULL_HANDLE;
    rt_frame_data *frame       = rtGetFrameData(g_gpu.current_frame_id);
    wait_semaphore             = frame->render_finished;

    uint32_t dummy = 0;
    if (rtSubmitSingleCommandBuffer(cmd,
                                    &wait_semaphore,
                                    &dummy,
                                    1,
                                    &transfer->ownership_transfer,
                                    &dummy,
                                    1,
                                    current_owner,
                                    VK_NULL_HANDLE) != RT_SUCCESS)
        return TRANSFER_FAILED;
    return TRANSFER_STARTED;
}

rt_result rtUploadToBuffer(VkBuffer buffer,
                           VmaAllocation allocation,
                           rt_gpu_queue current_owner,
                           const void *data,
                           size_t nbytes) {
    rtLockMutex(_transfer_lock);
    rt_transfer *transfer =
        (int)_transfer_count < rt_VkTransferSlotCount.i ? &_transfers[_transfer_count++] : NULL;
    rtUnlockMutex(_transfer_lock);
    if (!transfer)
        return RT_NO_TRANSFER_SLOTS;

    transfer->tbuf = AcquireTransferBuffer(nbytes);
    if (!transfer->tbuf.buffer) {
        return RT_OUT_OF_MEMORY;
    }

    CopyToTransferBuffer(transfer->tbuf, data, nbytes);

    if (!transfer->fence) {
        VkFenceCreateInfo fence_info = {
            .sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO,
        };
        if (vkCreateFence(g_gpu.device, &fence_info, g_gpu.alloc_cb, &transfer->fence) !=
            VK_SUCCESS) {
            return RT_UNKNOWN_ERROR;
        }
    }

    VkCommandBuffer cmd                 = rtAllocSingleCommandBuffer(RT_TRANSFER_QUEUE);
    VkCommandBufferBeginInfo begin_info = {.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
                                           .flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT};
    vkBeginCommandBuffer(cmd, &begin_info);

    bool requires_ownership_transfer =
        (current_owner != RT_TRANSFER_QUEUE && current_owner != RT_VK_UNOWNED);
    if (requires_ownership_transfer) {
        int did_transfer = AcquireBufferOwnership(transfer, buffer, current_owner, cmd);
        if (did_transfer == -1)
            return RT_UNKNOWN_ERROR;
        else if (did_transfer == TRANSFER_NOT_NEEDED)
            requires_ownership_transfer = false;
    }
    VkBufferCopy region = {.srcOffset = 0, .dstOffset = 0, .size = nbytes};
    vkCmdCopyBuffer(cmd, transfer->tbuf.buffer, buffer, 1, &region);
    vkEndCommandBuffer(cmd);

    uint32_t dummy = 0;
    return rtSubmitSingleCommandBuffer(cmd,
                                       requires_ownership_transfer ? &transfer->ownership_transfer
                                                                   : NULL,
                                       requires_ownership_transfer ? &dummy : NULL,
                                       requires_ownership_transfer ? 1 : 0,
                                       NULL,
                                       NULL,
                                       0,
                                       RT_TRANSFER_QUEUE,
                                       transfer->fence);
}

/* Wait until transfers to gpu resources are finished. */
void rtFlushGPUTransfers(void) {
    if (_transfer_count == 0)
        return;
    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);
    RT_ASSERT(temp.arena, "Could not get a temporary arena for flushing gpu transfers.");
    rtLockMutex(_transfer_lock);
    VkFence *fences = RT_ARENA_PUSH_ARRAY(temp.arena, VkFence, _transfer_count);
    if (!fences) {
        rtReportError("vk", "Failed to allocate fences array for flushing gpu transfers.");
        rtUnlockMutex(_transfer_lock);
        return;
    }
    uint32_t count = 0;
    for (uint32_t i = 0; i < _transfer_count; ++i) {
        if (!_transfers[i].fence || !_transfers[i].tbuf.buffer)
            continue;
        fences[count++] = _transfers[i].fence;
    }
    vkWaitForFences(g_gpu.device, count, fences, VK_TRUE, UINT64_MAX);
    for (uint32_t i = 0; i < _transfer_count; ++i) {
        if (!_transfers[i].fence || !_transfers[i].tbuf.buffer)
            continue;
        ReturnTransferBuffer(_transfers[i].tbuf);
    }
    _transfer_count = 0;
    rtUnlockMutex(_transfer_lock);
}
